
/* Written by Kristoffer Robin Stokke <krisrst@ifi.uio.no>
 *
 * For teaching purposes. Please don't hesitate to provide
 * me feedback, in case you have some constructive feedback
 * for the next years' students! */

#include <stdio.h>
#include <stdint.h>
#include <arm_neon.h>

/* Forward declare to gpu compute function */
/* nvcc doesn't handle arm_neon.h too well */
void gpu_matrix_transpose(uint8_t * src1, uint8_t * dst);

/* Writes matrix from memory */
void wr_matrix_8x8(uint8_t *vector){
	int i, j;

	for(i=0; i<8; i++){
		for(j=0; j<8; j++){
			fprintf(stdout, "%8d,", vector[i*8 + j]);
		}
		fprintf(stdout, "\n");
	}

	fprintf(stdout, "\n\n");
}

void transpose_8x8(uint8_t * matrix_1, uint8_t * matrix_out)
{
    int i, j;
    for(i = 0; i < 8; i ++)
        for(j = 0; j < 8; j ++)
        {
              matrix_out[i*8 + j] = matrix_1[j*8 + i]; 
        }
}

void simd_transpose_8x8(uint8_t * matrix_1, uint8_t * matrix_out)
{
    int i;
    uint8_t offset[16];
    uint8_t shift_values[16];

    // shift values = {0, 8, 16, 24 ... 56, 1, 9, 17, 25, ... 57}
    for(i=0; i<8; i++) 
    {
        offset[i]         = 2;
        offset[i+8]       = 2;
        shift_values[i]   =  i * 8;
        shift_values[i+8] = (i * 8) + 1;
    }

    __asm__ (
            "ld1 {v5.16b}, [%2]\n\t"                            // Set offset register
            "ld1 {v0.16b, v1.16b, v2.16b, v3.16b}, [%0]\n\t"    // Load source matrix
            "ld1 {v4.16b}, [%1]\n\t"                            // Load shift values

            /* V6 -> V9 : output matrix */
            "tbl v6.16b, {v0.16b, v1.16b, v2.16b, v3.16b}, v4.16b\n\t" // Shift elements from source
            "add v4.16b, v4.16b, v5.16b\n\t" // Increase offset vector
            "tbl v7.16b, {v0.16b, v1.16b, v2.16b, v3.16b}, v4.16b\n\t" // Repeat
            "add v4.16b, v4.16b, v5.16b\n\t"
            "tbl v8.16b, {v0.16b, v1.16b, v2.16b, v3.16b}, v4.16b\n\t"
            "add v4.16b, v4.16b, v5.16b\n\t"
            "tbl v9.16b, {v0.16b, v1.16b, v2.16b, v3.16b}, v4.16b\n\t"
            "add v4.16b, v4.16b, v5.16b\n\t"

            /* Store output vector */
            "st1 {v6.16b, v7.16b, v8.16b, v9.16b}, [%3]\n\t"
            :
            :"r" (matrix_1), "r" (shift_values), "r" (offset), "r" (matrix_out)
            : "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "memory"
            );
}

/* "Hardcore" version. Uses assembly for matrix multiplication */
void simd_mmul_8x8_hardcore(uint8_t *matrix_1, uint8_t *matrix_2, uint8_t *matrix_out){

	int i;

	for(i=0; i<8; i++){
		uint8_t *m1;
		uint8_t *m2;
		uint8_t *mo;

		m1 = matrix_1 + i*8;
		m2 = matrix_2 + i*8;
		mo = matrix_out + i*8;

		__asm__ (
			"ld1 {v0.8b}, [%1]\n\t"
			"ld1 {v1.8b}, [%2]\n\t"
			"mul v0.8b, v0.8b, v1.8b\n\t"
			"st1 {v0.8b}, [%0]\n\t"
		:
		:"r" (mo), "r" (m1), "r" (m2)
		: "v0", "v1", "memory"
		);
	}
}

/* Intrinsic version of matrix multiplication */
void simd_mmul_8x8_easy(uint8_t *matrix_1, uint8_t *matrix_2, uint8_t *matrix_out){

	int i;
	uint8x8_t m_1, m_2, m_o;

	for(i=0; i<8; i++){
		m_1 = vld1_u8(matrix_1 + i*8);
		m_2 = vld1_u8(matrix_2 + i*8);

		m_o = vmul_u8(m_1, m_2);

		vst1_u8(matrix_out + i*8, m_o);
	}
}

int main(void){

	fprintf(stdout, "NEON tutorial\n");

	int i, j;

	uint8_t matrix_1[8*8];
	uint8_t matrix_2[8*8];

	uint8_t matrix_out[8*8];

	for(j=0; j<8; j++){
		for(i=0; i<8; i++){
			matrix_1[j*8 + i] = i;
			matrix_2[j*8 + i] = i;
		}
	}

	fprintf(stdout, "First matrix:\n");
	wr_matrix_8x8(matrix_1);
//	fprintf(stdout, "Second matrix:\n");
//	wr_matrix_8x8(matrix_2);

	for(i=0; i<1E3; i++)
	{
        /*
		transpose_8x8(matrix_1, matrix_out);
		simd_mmul_8x8_easy(matrix_1, matrix_2, matrix_out);
		simd_mmul_8x8_hardcore(matrix_1, matrix_2, matrix_out);
		simd_transpose_8x8(matrix_1, matrix_out);
        */
		gpu_matrix_transpose(matrix_1, matrix_out);
	}

	(void) matrix_out;

	fprintf(stdout, "Resulting matrix:\n");
	wr_matrix_8x8(matrix_out);

	return 0;
}



